"""Authentication helper functions
"""
import os
from functools import wraps
from flask import request, Response
from .jwt import jwt_manager
from flask_jwt_extended import jwt_optional, get_jwt_identity


# Accept user name passed in Basic Auth header
# (password has to checkded before!)
ALLOW_BASIC_AUTH_USER = os.environ.get('ALLOW_BASIC_AUTH_USER', "False")
if ALLOW_BASIC_AUTH_USER.lower() in ("f", "false"):
    ALLOW_BASIC_AUTH_USER = False


def auth_manager(app, api=None):
    """Authentication setup for Flask app"""
    # Setup the Flask-JWT-Extended extension
    return jwt_manager(app, api)


def optional_auth(fn):
    """Authentication view decorator"""
    return jwt_optional(fn)


def get_auth_user():
    """Get username in authenticated request"""
    username = get_jwt_identity()
    if not username and ALLOW_BASIC_AUTH_USER:
        auth = request.authorization
        if auth:
            # We don't check password. Already done by WAF.
            username = auth.username
    return username



import os
import datetime
from flask_jwt_extended import JWTManager, unset_jwt_cookies
from flask import make_response


def jwt_manager(app, api=None):
    """Setup the Flask-JWT-Extended extension"""
    # http://flask-jwt-extended.readthedocs.io/en/latest/options.html
    app.config['JWT_TOKEN_LOCATION'] = ['headers', 'cookies']
    app.config['JWT_COOKIE_SECURE'] = False
    app.config['JWT_COOKIE_CSRF_PROTECT'] = False
    app.config['JWT_ACCESS_TOKEN_EXPIRES'] = datetime.timedelta(hours=12)
    app.config['JWT_SECRET_KEY'] = os.environ.get(
        'JWT_SECRET_KEY',
        'CHANGE-ME-1ef43ade8807dc37a6588cb8fb9dec4caf6dfd0e00398f9a')

    jwt = JWTManager(app)

    if api:
        # Delegate error handlers to RestPlus because of
        # https://github.com/vimalloc/flask-jwt-extended/issues/86
        jwt._set_error_handler_callbacks(api)

        @api.errorhandler
        def restplus_error_handler(error):
            # JWT error handler will be called afterwards
            return {}

    @jwt.expired_token_loader
    def handle_expired_token():
        # http://flask-jwt-extended.readthedocs.io/en/latest/changing_default_behavior.html
        # resp = redirect('/auth/login')
        # Automatic re-login does't work with SAML, so we prepare
        # for manual re-login
        resp = make_response()
        unset_jwt_cookies(resp)
        return resp

    return jwt



from datetime import datetime
import os
import time

import requests

from .cache import Cache


class PermissionClient():
    """PermissionClient class

    Get service specific permissions from permission service.
    """

    def __init__(self):
        """Constructor"""
        # get permission service URL from ENV (default: local service)
        self.service_url = os.environ.get(
            'CONFIG_SERVICE_URL', 'http://localhost:5010/').rstrip('/') + '/'
        # check if cache is valid every x seconds (default: 60s)
        self.config_check_interval = os.environ.get(
            'CONFIG_CHECK_INTERVAL', 60)
        # time in seconds until cache expiry (default: 24h)
        self.default_cache_duration = os.environ.get(
            'DEFAULT_CONFIG_CACHE_DURATION', 86400)

        self.headers = {
            'accept': 'application/json'
        }

        self.cache = Cache()
        self.last_update_check = None
        self.last_cache_flush = None

    def resource_permissions(self, resource_type, identity,
                             name=None, parent_id=None):
        """Return permitted resources for a resource type.

        :param str resource_type: Resource type
        :param obj identity: User name or Identity dict
        :param str name: Optional name filter
        :param str parent_id: Optional parent_id filter
        """
        params = {
            'resource_type': resource_type,
            'name': name,
            'parent_id': parent_id
        }
        return self.query_permissions(
            'resource_permissions', 'permissions/%s' % resource_type, params,
            identity, self.default_cache_duration, convert_int_keys=True
        )

    def resource_restrictions(self, resource_type, identity,
                              name=None, parent_id=None):
        """Return restricted resources for a resource type.

        :param str resource_type: Resource type
        :param obj identity: User name or Identity dict
        :param str name: Optional name filter
        :param str parent_id: Optional parent_id filter
        """
        params = {
            'resource_type': resource_type,
            'name': name,
            'parent_id': parent_id
        }
        return self.query_permissions(
            'resource_restrictions', 'restrictions/%s' % resource_type, params,
            identity, self.default_cache_duration, 'restrictions', True
        )

    def query_permissions(self, cache_key, path, params, identity,
                          cache_duration=86400, response_key='permissions',
                          convert_int_keys=False):
        """Return permissions or restrictions for a service or resource type.

        :param str cache_key: Key for cache lookup
        :param str path: Path for permissions ervice request
        :param obj params: Query specific request parameters
        :param obj identity: User name or Identity dict
        :param int cache_duration: Time in seconds until expiry (default: 24h)
        :param str response_key: Permissions key in JSON response
                                 (default: 'permissions')
        :paras bool convert_int_keys: Convert JSON integer keys to int
        """
        permissions = None
        if cache_duration:  # Caching active
            # check if cache is still valid
            self.check_cache()

            # get permissions from cache
            permissions = self.cache.read(
                cache_key, identity, params.values())
            if permissions:
                return permissions

        # build request URL: http://<service_url>/<path>
        url = self.service_url + path

        reqparams = params.copy()  # don't change params before cache.write
        if identity:
            if isinstance(identity, dict):
                reqparams.update({
                    'username': identity.get('username'),
                    'group': identity.get('group')
                })
            else:
                # identity is username
                reqparams.update({'username': identity})

        # send request to permission service
        response = requests.get(url, headers=self.headers, params=reqparams,
                                timeout=30)
        if response.status_code == requests.codes.ok:
            if convert_int_keys:
                permissions = {}
                result = response.json()[response_key]
                for key, value in result.items():
                    # convert integer keys from JSON from string to int
                    try:
                        key = int(key)
                    except ValueError:
                        pass
                    permissions[key] = value
            else:
                permissions = response.json()[response_key]

        if cache_duration:  # Caching active
            self.cache.write(cache_key, identity, params.values(),
                             permissions, cache_duration)

        return permissions

    def dataset_edit_permissions(self, dataset, identity):
        """Return dataset edit permissions if available and permitted.

        :param str dataset: Dataset ID
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'data', 'data', {'dataset': dataset}, identity,
            self.default_cache_duration
        )

    def document_permissions(self, template, identity):
        """Return document template permissions if available and permitted.

        :param str dataset: Template ID
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'document', 'document', {'template': template}, identity,
            self.default_cache_duration
        )

    def feature_info_permissions(self, ows_name, identity):
        """Return feature info template permissions if available and permitted.

        :param str ows_name: WMS service name
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'feature_info', 'feature_info', {'ows_name': ows_name}, identity,
            self.default_cache_duration
        )

    def ogc_permissions(self, ows_name, ows_type, identity):
        """Return OGC service permissions if available and permitted.

        :param str ows_name: OWS service name
        :param str ows_type: OWS type (WMS or WFS)
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'ogc', 'ogc', {'ows_name': ows_name, 'ows_type': ows_type},
            identity, self.default_cache_duration
        )

    def print_permissions(self, template, identity):
        """Return print template permissions if available and permitted.

        :param str dataset: Template ID
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'print', 'print', {'template': template}, identity,
            self.default_cache_duration
        )

    def qwc_permissions(self, identity, viewer=None):
        """Return data for QWC themes.json for available and permitted
        resources.

        :param obj identity: User name or Identity dict
        :param str viewer: Optional custom viewer name (None for default)
        """
        return self.query_permissions(
            'qwc', 'qwc', {'viewer': viewer}, identity,
            self.default_cache_duration
        )

    def dataset_search_permissions(self, dataset, identity):
        """Return dataset search permissions if available and permitted.

        :param str dataset: Dataset ID
        :param obj identity: User name or Identity dict
        """
        return self.query_permissions(
            'search', 'search', {'dataset': dataset}, identity,
            self.default_cache_duration
        )

    def check_cache(self):
        """Check if cache is still valid and flush if obsolete."""
        if (self.last_update_check is None
            or self.last_update_check +
                self.config_check_interval < time.time()):
            # get last permissions update from permission service
            url = self.service_url + 'last_update'
            response = requests.get(url, headers=self.headers, timeout=30)
            if response.status_code == requests.codes.ok:
                permissions_updated_at = datetime.strptime(
                    response.json()['permissions_updated_at'],
                    "%Y-%m-%d %H:%M:%S"
                )
                if (self.last_cache_flush is None
                        or self.last_cache_flush < permissions_updated_at):
                    # flush cache if obsolete
                    self.cache.init()
                    self.last_cache_flush = permissions_updated_at

            self.last_update_check = time.time()



import os
from flask import json, safe_join


class PermissionsReader():
    """PermissionsReader helper class

    Read users, groups, roles and permissions for a tenant from a JSON file.
    Provides helper methods for collectiong permissions.
    """

    # name of public role
    PUBLIC_ROLE_NAME = 'public'

    @staticmethod
    def permissions_file_path(tenant):
        """Return path to permissions JSON file for a tenant.

        :param str tenant: Tenant ID
        """
        config_path = os.environ.get('CONFIG_PATH', 'config')
        return safe_join(config_path, tenant, 'permissions.json')

    def __init__(self, tenant, logger):
        """Constructor

        :param str tenant: Tenant ID
        :param Logger logger: Application logger
        """
        self.tenant = tenant
        self.logger = logger
        self.permissions = self.load_permissions()

    def read_permissions(self):
        """Read permissions for a tenant from a JSON file."""
        permissions = {}

        permissions_path = PermissionsReader.permissions_file_path(self.tenant)
        self.logger.info("Reading permissions '%s'" % permissions_path)
        try:
            with open(permissions_path, encoding='utf-8') as fh:
                permissions = json.load(fh)
        except Exception as e:
            self.logger.error(
                "Could not load permissions '%s':\n%s" %
                (permissions_path, e)
            )
            raise e
        # TODO: validate permissions schema

        return permissions

    def load_permissions(self):
        """Load users, groups, roles and permissions.

        Returns lookup dict as:
            {
                users: {
                    <user>: [<role>]
                }
                groups: {
                    <group>: [<role>]
                },
                roles: {
                    <role>: <permissions{}>
                }
            }
        """
        permissions = self.read_permissions()

        # transform raw permissions to lookup dict

        # collect group roles
        groups = {}
        for group in permissions.get('groups', []):
            groups[group['name']] = group['roles']

        # collect user roles
        users = {}
        for user in permissions.get('users', []):
            user_roles = []
            # direct roles
            for role in user['roles']:
                user_roles.append(role)
            # roles from groups
            for group in user['groups']:
                user_roles.extend(groups.get(group, []))
            # assign unique sorted roles
            users[user['name']] = sorted(list(set(user_roles)))

        # collect role permissions
        roles = {}
        for role in permissions.get('roles', []):
            roles[role['role']] = role['permissions']

        return {
            'users': users,
            'groups': groups,
            'roles': roles
        }

    # helpers

    def identity_roles(self, identity):
        """Return roles for identity.

        :param obj identity: User identity
        """
        # extract username and group
        username = None
        group = None
        if identity:
            if isinstance(identity, dict):
                username = identity.get('username')
                group = identity.get('group')
            else:
                # identity is username
                username = identity

        # add default public role
        roles = [self.PUBLIC_ROLE_NAME]
        # add any user roles
        roles.extend(self.permissions['users'].get(username, []))
        # add any group roles
        roles.extend(self.permissions['groups'].get(group, []))

        # return unique sorted roles
        return sorted(list(set(roles)))

    def resource_permissions(self, resource_key, identity, resource_name=None):
        """Return collected list of resource permissions for identity roles.

        :param str resource_key: Resource key in permissions data
        :param obj identity: User identity
        :param str name: Optional resource name filter
        """
        permissions = []

        roles = self.identity_roles(identity)
        for role in roles:
            # get role permissions
            role_permissions = self.permissions['roles'].get(role, {})
            # get permissions for resource key
            resource_permissions = role_permissions.get(resource_key, {})
            if resource_name is not None:
                # filter by resource name
                for permission in resource_permissions:
                    if isinstance(permission, dict):
                        if permission.get('name') == resource_name:
                            permissions.append(permission)
                    else:
                        if permission == resource_name:
                            permissions.append(permission)
            else:
                permissions.extend(resource_permissions)

        return permissions



from datetime import datetime
import os
import re
from flask import request

from .permissions_reader import PermissionsReader
from .runtime_config import RuntimeConfig


DEFAULT_TENANT = 'default'


class TenantHandler:
    """Tenant handling class
    """

    def __init__(self, logger):
        """Constructor

        :param Logger logger: Application logger
        """
        self.logger = logger
        self.tenant_name = os.environ.get('QWC_TENANT')
        self.tenant_url_re = os.environ.get('TENANT_URL_RE')
        if self.tenant_url_re:
            self.tenant_url_re = re.compile(self.tenant_url_re)
        self.tenant_referrer_re = os.environ.get('TENANT_REFERRER_RE')
        if self.tenant_referrer_re:
            self.tenant_referrer_re = re.compile(self.tenant_referrer_re)
        self.handler_cache = {}  # handler_cache[handler_name][tenant]

    def tenant(self, identity):
        if self.tenant_name:
            return self.tenant_name
        if self.tenant_url_re:
            # self.logger.debug("Extracting tenant from base_url %s" %
            #                   request.base_url)
            match = self.tenant_url_re.match(request.base_url)
            if match:
                return match.group(1)
            else:
                return DEFAULT_TENANT
        if self.tenant_referrer_re and request.referrer:
            # self.logger.debug("Extracting tenant from referrer URL %s" %
            #                   request.referrer)
            match = self.tenant_referrer_re.match(request.referrer)
            if match:
                return match.group(1)
        return DEFAULT_TENANT

    def handler(self, service_name, handler_name, tenant):
        """Get service handler for tenant.

        Return None if not yet registered or if config files have changed.

        :param str service_name: Service name
                                 (used for detecting config changes)
        :param str handler_name: Handler name
        :param str tenant: Tenant ID
        """
        handlers = self.handler_cache.get(handler_name)
        if handlers:
            handler = handlers.get(tenant)
            if handler:
                # check for config updates
                last_update = self.last_config_update(service_name, tenant)
                if last_update and last_update < handler.get('last_update'):
                    # cache is up-to-date
                    return handler.get('handler')
                else:
                    # config has changed, remove handler from cache
                    del handlers[tenant]

        return None

    def register_handler(self, handler_name, tenant, handler):
        """Register service handler for tenant"""
        handlers = self.handler_cache.get(handler_name)
        if handlers is None:
            handlers = {}
            self.handler_cache[handler_name] = handlers
        handlers[tenant] = {
            'handler': handler,
            'last_update': datetime.utcnow()
        }
        return handler

    def last_config_update(self, service_name, tenant):
        """Return latest timestamp of config and permission files for a tenant.

        :param str service_name: Service name
        :param str tenant: Tenant ID
        """
        # get latest timestamp of config and permission files
        last_config_update = None
        paths = [
            RuntimeConfig.config_file_path(service_name, tenant),
            PermissionsReader.permissions_file_path(tenant)
        ]
        for path in paths:
            if os.path.isfile(path):
                timestamp = datetime.utcfromtimestamp(
                    os.path.getmtime(path)
                )
                if (
                    last_config_update is None
                    or timestamp > last_config_update
                ):
                    last_config_update = timestamp

        return last_config_update
